from functools import wraps

from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.core.exceptions import PermissionDenied, ObjectDoesNotExist
from django.shortcuts import resolve_url
from django.contrib.contenttypes.models import ContentType
from django.utils.decorators import available_attrs
from django.utils.six.moves.urllib.parse import urlparse
from django.http import Http404
from operator import and_, or_

from trusts import utils


class P(object):
    def __init__(self, perm, **fieldlookups):
        self._perm = perm
        self._fieldlookups = fieldlookups
        self._left_operand = None
        self._right_operand = None
        self._operator = None

    def __and__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError("unsupported operand type(s) for &: '%s' and '%s'" % type(self), type(other))

        p = type(self)('')
        p._left_operand = self
        p._right_operand = other
        p._operator = and_
        return p

    def __or__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError("unsupported operand type(s) for |: '%s' and '%s'" % type(self), type(other))

        p = type(self)('')
        p._left_operand = self
        p._right_operand = other
        p._operator = or_
        return p

    def __repr__(self):
        return self.__unicode__()

    def __unicode__(self):
        if not self._operator:
            return self.perm

        return 'P object'

    def get_leaves(self):
        leaves = []
        if not self._operator:
            return [self]

        # Do not use += or leaves.extend here since it changes the original list
        leaves = leaves + self._left_operand.get_leaves()
        leaves = leaves + self._right_operand.get_leaves()

        return leaves

    def solve(self, fn):
        if self._operator:
            # Parent node, return result operation
            if self._operator == and_:
                return self._left_operand.solve(fn) and self._right_operand.solve(fn)
            elif self._operator == or_:
                return self._left_operand.solve(fn) or self._right_operand.solve(fn)
            else:
                raise TypeError('Unsupported Operator: ', self._operator)
        else:
            return fn(self._perm, **self._fieldlookups)


class R(object):
    def __init__(self, key):
        self.key = key


class K(R):
    pass


class G(R):
    pass


class O(R):
    pass


def request_passes_test(test_func, login_url=None, redirect_field_name=REDIRECT_FIELD_NAME, *args, **kwargs):
    '''
    Decorator for views that checks that the user passes the given test,
    redirecting to the log-in page if necessary. The test should be a callable
    that takes the user object and returns True if the user passes.

    Adapted from `django/contrib/auth/decorator.py`
    '''

    def decorator(view_func):
        @wraps(view_func, assigned=available_attrs(view_func))
        def _wrapped_view(request, *args, **kwargs):
            if test_func(request, *args, **kwargs):
                return view_func(request, *args, **kwargs)

            path = request.build_absolute_uri()
            resolved_login_url = resolve_url(login_url or settings.LOGIN_URL)
            # If the login url is the same scheme and net location then just
            # use the path as the "next" url.
            login_scheme, login_netloc = urlparse(resolved_login_url)[:2]
            current_scheme, current_netloc = urlparse(path)[:2]
            if ((not login_scheme or login_scheme == current_scheme) and
                    (not login_netloc or login_netloc == current_netloc)):
                path = request.get_full_path()
            from django.contrib.auth.views import redirect_to_login
            return redirect_to_login(
                path, resolved_login_url, redirect_field_name)
        return _wrapped_view
    return decorator


def _collect_args(args, fieldlookups):
    results = {}

    if not fieldlookups:
        return results

    for lookup, arg_name in fieldlookups.iteritems():
        if arg_name in args:
            results[lookup] = args[arg_name]
        else:
            results[lookup] = None

    return results


def _get_permissible_items(perm, request, fieldlookups):
    if fieldlookups is None:
        return None

    applabel, modelname, action, cond = utils.parse_perm_code(perm)
    try:
        ctype = ContentType.objects.get_by_natural_key(applabel, modelname)

        return ctype.model_class().objects.filter(**fieldlookups)
    except ObjectDoesNotExist:
        raise ValueError('Permission code must be of the form "app_label.action_modelname". Actual: %s' % permext)


def _resolve_fieldlookups(request, kwargs, fieldlookups_kwargs=None, fieldlookups_getparams=None, fieldlookups_postparams=None, **fieldlookups):
    resolved_fields = {}

    resolved_fields.update(_collect_args(kwargs, fieldlookups_kwargs))
    resolved_fields.update(_collect_args(request.GET, fieldlookups_getparams))
    resolved_fields.update(_collect_args(request.POST, fieldlookups_postparams, ))

    for field, lookup in fieldlookups.items():
        if isinstance(lookup, K):
            source = kwargs
        elif isinstance(lookup, G):
            source = request.GET
        elif isinstance(lookup, O):
            source = request.POST
        else:
            continue
        resolved_fields[field] = source[lookup.key] if lookup.key in source else None

    return resolved_fields or None


def _check(perm, request, kwargs, raise_exception, **fieldlookups):
    if not isinstance(perm, (list, tuple)):
        perms = (perm, )
    else:
        perms = perm

    resolved_items = _resolve_fieldlookups(request, kwargs, **fieldlookups)

    items = None
    if fieldlookups is not None:
        items = _get_permissible_items(perm, request, resolved_items)
        if items is None:
            if raise_exception:
                raise Http404
            return False

    if request.user.has_perms(perms, items):
        return True

    # In case the 403 handler should be called raise the exception
    if raise_exception:
        raise PermissionDenied
    return False


def permission_required(perm, raise_exception=True, login_url=None, **fieldlookups):
    '''
    Decorator for views that checks whether a user has a particular permission
    enabled, redirecting to the log-in page if necessary.
    If the raise_exception parameter is given the PermissionDenied exception
    is raised.

    Adapted from `django/contrib/auth/decorator.py`
    '''

    def _check_perms(request, *args, **kwargs):
        def _wrapped_check(perm, **fieldlookups):
            return _check(perm, request, kwargs, raise_exception, **fieldlookups)

        if isinstance(perm, P):
            return perm.solve(_wrapped_check)

        return _check(perm, request, kwargs, raise_exception, **fieldlookups)

    return request_passes_test(_check_perms, login_url=login_url)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = 'trusts'
    verbose_name = "Django Trusts Add-in"
    label = 'trusts'



# see https://docs.djangoproject.com/en/1.9/topics/testing/advanced/#using-the-django-test-runner-to-test-reusable-applications
import os, sys
os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.settings'
test_dir = os.path.join(os.path.dirname(__file__), 'trusts')
sys.path.insert(0, test_dir)

import django
from django.test.utils import get_runner
from django.conf import settings

def runtests():
    TestRunner = get_runner(settings)
    test_runner = TestRunner(verbosity=1, interactive=True)
    if hasattr(django, 'setup'):
        django.setup()
    failures = test_runner.run_tests(['trusts'])
    sys.exit(bool(failures))

if __name__ == '__main__':
    runtests()



from __future__ import unicode_literals

from django.db.models import F, Q, QuerySet
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend

from trusts.models import Trust, Content
from trusts import get_permission_model, utils


class TrustModelBackendMixin(object):
    perm_model = get_permission_model()

    @staticmethod
    def _get_perm_code(perm):
        return '%s.%s' % (
            perm.content_type.app_label, perm.codename
        )

    @staticmethod
    def _get_trusts(obj):
        if not Content.is_content(obj):
            return []

        trusts = Trust.objects.filter_by_content(obj)
        if trusts is None:
            return []

        if not hasattr(trusts, '__iter__'):
            trusts = [trusts]

        return trusts

    @staticmethod
    def _get_class(obj):
        if isinstance(obj, QuerySet):
            klass = obj.model
        else:
            klass = obj.__class__
        return klass

    def get_group_permissions(self, user_obj, obj=None):
        """
        Returns a set of permission strings that this user has through his/her
        groups.
        """

        if user_obj.is_anonymous() or obj is None:
            return super(TrustModelBackendMixin, self).get_group_permissions(user_obj, obj)

        if Content.is_content(obj):
            filter = self.perm_model.objects.filter
            return filter(group__trusts__in=self._get_trusts(obj), group__user=user_obj)

        return []

    def get_all_permissions(self, user_obj, obj=None):
        if user_obj.is_anonymous() or obj is None:
            return super(TrustModelBackendMixin, self).get_all_permissions(user_obj, obj)

        if not hasattr(user_obj, '_trust_perm_cache'):
            setattr(user_obj, '_trust_perm_cache', dict())
        perm_cache = getattr(user_obj, '_trust_perm_cache')

        trusts = self._get_trusts(obj)
        if len(trusts):
            all_perms = []
            for trust in trusts:
                if trust.pk not in perm_cache.keys():
                    trust_perm = set([self._get_perm_code(p) for p in
                        self.perm_model.objects.filter(
                            Q(group__trusts=trust, group__user=user_obj) |
                            Q(roles__groups__trusts=trust, roles__groups__user=user_obj) |
                            Q(trustentities__trust=trust, trustentities__entity=user_obj)
                        )
                        .order_by('group__trusts', 'trustentities__entity')
                    ])

                    perm_cache[trust.pk] = trust_perm
                else:
                    trust_perm = perm_cache[trust.pk]

                all_perms.append(trust_perm)
            return set.intersection(*all_perms)
        return []

    def permission_condition_met(self, func, user_obj, perm, obj):
        if isinstance(obj, QuerySet):
            objs = obj.all()
        elif hasattr(trusts, '__iter__'):
            objs = obj
        else:
            objs = [obj]

        return all([func(user_obj, perm, o) for o in objs])

    def has_perm(self, user_obj, permext, obj=None):
        applabel, modelname, action, cond = utils.parse_perm_code(permext)
        if len(cond) != 0:
            func = Content.get_permission_condition_func(self._get_class(obj), cond)
            if func is None:
                raise AttributeError('Permission condition code "%s" is not associate with model "%s_%s"' % (cond, applabel, modelname))

        perm = '%s.%s_%s' % (applabel, action, modelname)
        positive = super(TrustModelBackendMixin, self).has_perm(user_obj=user_obj, perm=perm, obj=obj)
        if positive:
            if len(cond) == 0:
                return True

            if self.permission_condition_met(func, user_obj, perm, obj):
                return True
        return False


class TrustModelBackend(TrustModelBackendMixin, ModelBackend):
    pass



from django.contrib import admin
from trusts.models import Trust, Role, RolePermission, TrustUserPermission

admin.site.register(Trust, admin.ModelAdmin)
admin.site.register(Role, admin.ModelAdmin)
admin.site.register(RolePermission, admin.ModelAdmin)
admin.site.register(TrustUserPermission, admin.ModelAdmin)




SECRET_KEY = '01)%8q7ub=+yw7^#dz5s!6kkff6%al5f)_ayvep9_b&w1q-dvs'

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'trusts',
)

AUTHENTICATION_BACKENDS = (
    'trusts.backends.TrustModelBackend',
)
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

ROOT_URLCONF = 'tests.urls'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import os
from setuptools import setup, find_packages


README = open(os.path.join(os.path.dirname(__file__), 'README.md')).read()
REQUIREMENTS = open(os.path.join(os.path.dirname(__file__), 'requirements.txt')).read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-trusts',
    version='0.10.3',
    description='Django authorization add-on for multiple organizations and object-level permission settings',
    author='Thomas Yip',
    author_email='thomasleaf@gmail.com',
    long_description=README,
    url='http://github.com/beedesk/django-trusts',
    packages=find_packages(exclude=[]),
    test_suite="tests.runtests.runtests",
    include_package_data=True,
    zip_safe=False,
    install_requires=REQUIREMENTS,
    license='BSD 2-Clause',
    classifiers=[
	'Development Status :: 4 - Beta',
	'Framework :: Django :: 1.8',
	'Intended Audience :: Developers',
	'Intended Audience :: Information Technology',
	'Intended Audience :: System Administrators',
	'License :: OSI Approved :: BSD License',
	'Natural Language :: English',
	'Operating System :: OS Independent',
	'Programming Language :: Python',
	'Programming Language :: Python :: 2',
	'Topic :: Software Development :: Libraries',
    ],
)
